#include "opencv2/opencv.hpp"

int getLabelValue(const cv::Mat& img, int row, int col)
{
    if (col < 0 || row < 0 || col >= img.cols || row >= img.rows) {
        return 0;
    }
    return img.at<int>(row, col);
}

void updateConnection(std::vector<int>& con,
                      int label1, int label2, int cur_label)
{
    // XXX: can be faster

    int p1 = con[label1];
    int p2 = con[label2];

    if (p1 == p2) {
        return;
    }
    if (p1 > p2) { // swap p1 and p2
        int t = p1;
        p1 = p2;
        p2 = t;
    }
    for (int i = 0; i <= cur_label; i++) {
        if (con[i] == p2) {
            con[i] = p1;
        }
    }
}

int main()
{
    cv::namedWindow("input", cv::WINDOW_AUTOSIZE);
    cv::namedWindow("output", cv::WINDOW_AUTOSIZE);

    cv::VideoCapture cap(0);
    cv::Mat input;
    cap >> input;
    cv::Mat output(input.rows, input.cols, CV_8U);
    cv::Mat label(input.rows, input.cols, CV_32S);

    int thresh = 150, min_area = 20;
    cv::createTrackbar("threshold", "input", &thresh, 255);
    cv::createTrackbar("min area", "input", &min_area, 1023);

    const int max_label = INT_MAX;

    while (1) {

        std::vector<int> connection;
        connection.push_back(0); // label 0 is for 0-value pixels

        cap >> input;
        cv::cvtColor(input, input, CV_BGR2GRAY);
        cv::threshold(input, input, thresh, 255, cv::THRESH_BINARY);

        label = 0;
        int cur_label = 0; // largest label allocated so far
        int width = input.cols;
        int height = input.rows;
        
        for (int j = 0; j < height; j++) {
            for (int i = 0; i < width; i++) {
                if (input.at<uchar>(j, i) != 0) {

                    int u_label = getLabelValue(label, j - 1, i);
                    int l_label = getLabelValue(label, j, i - 1);
                    
                    if (u_label > 0 && l_label > 0) {
                        if (u_label == l_label) {
                            label.at<int>(j, i) = u_label;
                        } else if (u_label < l_label) {
                            label.at<int>(j, i) = u_label;
                            updateConnection(connection,
                                             u_label, l_label, cur_label);
                        } else {
                            label.at<int>(j, i) = l_label;
                            updateConnection(connection,
                                             l_label, u_label, cur_label);
                        }
                    } else if (u_label > 0 && l_label == 0) {
                        label.at<int>(j, i) = u_label;
                    } else if (u_label == 0 && l_label > 0) {
                        label.at<int>(j, i) = l_label;
                    } else if (cur_label < max_label) {
                        ++cur_label;
                        label.at<int>(j, i) = cur_label;
                        connection.push_back(cur_label);
                    } else { // exceed max labels
                        label.at<int>(j, i) = 0;
                    }
                }
            }
        }

        // re-number the labels (optional)
        int last_updated_label = 0;
        int cur_real_label = 0;
        for (int i = 1; i <= cur_label; i++) {
            if (connection[i] > last_updated_label) {
                last_updated_label = connection[i];
                connection[i] = ++cur_real_label;
            } else {
                connection[i] = connection[connection[i]];
            }
        }

        // generate label image and compute moments
        std::vector<int> m00(cur_real_label + 1);
        std::vector<int> m10(cur_real_label + 1);
        std::vector<int> m01(cur_real_label + 1);
        for (int j = 0; j < height; j++) {
            for (int i = 0; i < width; i++) {
                int real_label = connection[label.at<int>(j, i)];
                label.at<int>(j, i) = real_label;
                m00[real_label]++;
                m10[real_label] += i;
                m01[real_label] += j;
            }
        }

        // display the results
        cv::cvtColor(input, output, CV_GRAY2BGR);
        if (min_area == 0) {
            min_area = 1;
        }
        for (int i = 1; i <= cur_real_label; i++) {
            int gx, gy;
            if (m00[i] < min_area) {
                continue;
            }
            gx = m10[i] / m00[i];
            gy = m01[i] / m00[i];
            cv::putText(output, cv::format("%d", i), cv::Point(gx, gy),
                        CV_FONT_HERSHEY_PLAIN, 1.0, CV_RGB(255, 0, 0));
        }

        cv::imshow("input", input);
        cv::imshow("output", output);
        if (cv::waitKey(30) > 0) {
            break;
        }
    }
}
